const char load_cfg_lua[] =
"-- load_cfg.lua - internal file\n"
"\n"
"local log = require('log')\n"
"local json = require('json')\n"
"local private = require('box.internal')\n"
"local urilib = require('uri')\n"
"local math = require('math')\n"
"\n"
"-- all available options\n"
"local default_cfg = {\n"
"    listen              = nil,\n"
"    memtx_memory        = 256 * 1024 *1024,\n"
"    memtx_min_tuple_size = 16,\n"
"    memtx_max_tuple_size = 1024 * 1024,\n"
"    slab_alloc_factor   = 1.05,\n"
"    work_dir            = nil,\n"
"    memtx_dir           = \".\",\n"
"    wal_dir             = \".\",\n"
"\n"
"    vinyl_dir           = '.',\n"
"    vinyl_memory        = 128 * 1024 * 1024,\n"
"    vinyl_cache         = 128 * 1024 * 1024,\n"
"    vinyl_max_tuple_size = 1024 * 1024,\n"
"    vinyl_read_threads  = 1,\n"
"    vinyl_write_threads = 2,\n"
"    vinyl_timeout       = 60,\n"
"    vinyl_run_count_per_level = 2,\n"
"    vinyl_run_size_ratio      = 3.5,\n"
"    vinyl_range_size          = 1024 * 1024 * 1024,\n"
"    vinyl_page_size           = 8 * 1024,\n"
"    vinyl_bloom_fpr           = 0.05,\n"
"    log                 = nil,\n"
"    log_nonblock        = true,\n"
"    log_level           = 5,\n"
"    io_collect_interval = nil,\n"
"    readahead           = 16320,\n"
"    snap_io_rate_limit  = nil, -- no limit\n"
"    too_long_threshold  = 0.5,\n"
"    wal_mode            = \"write\",\n"
"    rows_per_wal        = 500000,\n"
"    wal_max_size        = 256 * 1024 * 1024,\n"
"    wal_dir_rescan_delay= 2,\n"
"    force_recovery      = false,\n"
"    replication         = nil,\n"
"    custom_proc_title   = nil,\n"
"    pid_file            = nil,\n"
"    background          = false,\n"
"    username            = nil,\n"
"    coredump            = false,\n"
"    read_only           = false,\n"
"    hot_standby         = false,\n"
"    checkpoint_interval = 3600,\n"
"    checkpoint_count    = 2,\n"
"    worker_pool_threads = 4,\n"
"    replication_timeout = 10,\n"
"}\n"
"\n"
"-- types of available options\n"
"-- could be comma separated lua types or 'any' if any type is allowed\n"
"local template_cfg = {\n"
"    listen              = 'string, number',\n"
"    memtx_memory        = 'number',\n"
"    memtx_min_tuple_size  = 'number',\n"
"    memtx_max_tuple_size  = 'number',\n"
"    slab_alloc_factor   = 'number',\n"
"    work_dir            = 'string',\n"
"    memtx_dir            = 'string',\n"
"    wal_dir             = 'string',\n"
"    vinyl_dir           = 'string',\n"
"    vinyl_memory        = 'number',\n"
"    vinyl_cache               = 'number',\n"
"    vinyl_max_tuple_size      = 'number',\n"
"    vinyl_read_threads        = 'number',\n"
"    vinyl_write_threads       = 'number',\n"
"    vinyl_timeout             = 'number',\n"
"    vinyl_run_count_per_level = 'number',\n"
"    vinyl_run_size_ratio      = 'number',\n"
"    vinyl_range_size          = 'number',\n"
"    vinyl_page_size           = 'number',\n"
"    vinyl_bloom_fpr           = 'number',\n"
"\n"
"    log              = 'string',\n"
"    log_nonblock     = 'boolean',\n"
"    log_level           = 'number',\n"
"    io_collect_interval = 'number',\n"
"    readahead           = 'number',\n"
"    snap_io_rate_limit  = 'number',\n"
"    too_long_threshold  = 'number',\n"
"    wal_mode            = 'string',\n"
"    rows_per_wal        = 'number',\n"
"    wal_max_size        = 'number',\n"
"    wal_dir_rescan_delay= 'number',\n"
"    force_recovery      = 'boolean',\n"
"    replication         = 'string, number, table',\n"
"    custom_proc_title   = 'string',\n"
"    pid_file            = 'string',\n"
"    background          = 'boolean',\n"
"    username            = 'string',\n"
"    coredump            = 'boolean',\n"
"    checkpoint_interval = 'number',\n"
"    checkpoint_count    = 'number',\n"
"    read_only           = 'boolean',\n"
"    hot_standby         = 'boolean',\n"
"    worker_pool_threads = 'number',\n"
"    replication_timeout = 'number',\n"
"}\n"
"\n"
"local function normalize_uri(port)\n"
"    if port == nil or type(port) == 'table' then\n"
"        return port\n"
"    end\n"
"    return tostring(port);\n"
"end\n"
"\n"
"-- options that require special handling\n"
"local modify_cfg = {\n"
"    listen             = normalize_uri,\n"
"    replication        = normalize_uri,\n"
"}\n"
"\n"
"local function purge_password_from_uri(uri)\n"
"    local parsed = urilib.parse(uri)\n"
"    if parsed ~= nil and parsed.password ~= nil then\n"
"        return urilib.format(parsed, false)\n"
"    end\n"
"    return uri\n"
"end\n"
"\n"
"local function purge_password_from_uris(uri)\n"
"    if uri == nil then\n"
"        return nil\n"
"    end\n"
"    if type(uri) == 'table' then\n"
"        local new_table = {}\n"
"        for k, v in pairs(uri) do\n"
"            new_table[k] = purge_password_from_uri(v)\n"
"        end\n"
"        return new_table\n"
"    end\n"
"    return purge_password_from_uri(uri)\n"
"end\n"
"\n"
"-- options that require modification for logging\n"
"local log_cfg_option = {\n"
"    replication = purge_password_from_uris,\n"
"}\n"
"\n"
"-- dynamically settable options\n"
"local dynamic_cfg = {\n"
"    listen                  = private.cfg_set_listen,\n"
"    replication             = private.cfg_set_replication,\n"
"    log_level               = private.cfg_set_log_level,\n"
"    io_collect_interval     = private.cfg_set_io_collect_interval,\n"
"    readahead               = private.cfg_set_readahead,\n"
"    too_long_threshold      = private.cfg_set_too_long_threshold,\n"
"    snap_io_rate_limit      = private.cfg_set_snap_io_rate_limit,\n"
"    read_only               = private.cfg_set_read_only,\n"
"    memtx_max_tuple_size    = private.cfg_set_memtx_max_tuple_size,\n"
"    vinyl_max_tuple_size    = private.cfg_set_vinyl_max_tuple_size,\n"
"    vinyl_timeout           = private.cfg_set_vinyl_timeout,\n"
"    checkpoint_count        = private.cfg_set_checkpoint_count,\n"
"    checkpoint_interval     = private.checkpoint_daemon.set_checkpoint_interval,\n"
"    worker_pool_threads     = private.cfg_set_worker_pool_threads,\n"
"    -- do nothing, affects new replicas, which query this value on start\n"
"    wal_dir_rescan_delay    = function() end,\n"
"    custom_proc_title       = function()\n"
"        require('title').update(box.cfg.custom_proc_title)\n"
"    end,\n"
"    force_recovery          = function() end,\n"
"    replication_timeout     = function() end,\n"
"}\n"
"\n"
"local dynamic_cfg_skip_at_load = {\n"
"    wal_mode                = true,\n"
"    listen                  = true,\n"
"    replication             = true,\n"
"    wal_dir_rescan_delay    = true,\n"
"    custom_proc_title       = true,\n"
"    force_recovery          = true,\n"
"}\n"
"\n"
"local function convert_gb(size)\n"
"    return math.floor(size * 1024 * 1024 * 1024)\n"
"end\n"
"\n"
"-- Old to new config translation tables\n"
"local translate_cfg = {\n"
"    snapshot_count = {'checkpoint_count'},\n"
"    snapshot_period = {'checkpoint_interval'},\n"
"    slab_alloc_arena = {'memtx_memory', convert_gb},\n"
"    slab_alloc_minimal = {'memtx_min_tuple_size'},\n"
"    slab_alloc_maximal = {'memtx_max_tuple_size'},\n"
"    snap_dir = {'memtx_dir'},\n"
"    logger = {'log'},\n"
"    logger_nonblock = {'log_nonblock'},\n"
"    panic_on_snap_error = {'force_recovery', function (p) return not p end},\n"
"    panic_on_wal_error = {'force_recovery', function (p) return not p end},\n"
"    replication_source = {'replication'},\n"
"}\n"
"\n"
"-- Upgrade old config\n"
"local function upgrade_cfg(cfg, translate_cfg)\n"
"    if cfg == nil then\n"
"        return {}\n"
"    end\n"
"    local result_cfg = {}\n"
"    for k, v in pairs(cfg) do\n"
"        local translation = translate_cfg[k]\n"
"        if translation ~= nil then\n"
"            log.warn('Deprecated option %s, please use %s instead', k, translation[1])\n"
"            local new_val\n"
"            if translation[2] == nil then\n"
"                new_val = v\n"
"            else\n"
"                new_val = translation[2](v)\n"
"            end\n"
"            if cfg[translation[1]] ~= nil and\n"
"               cfg[translation[1]] ~= new_val then\n"
"                box.error(box.error.CFG, k,\n"
"                          'can not override a value for a deprecated option')\n"
"            end\n"
"            result_cfg[translation[1]] = new_val\n"
"        else\n"
"            result_cfg[k] = v\n"
"        end\n"
"    end\n"
"    return result_cfg\n"
"end\n"
"\n"
"local function prepare_cfg(cfg, default_cfg, template_cfg, modify_cfg, prefix)\n"
"    if cfg == nil then\n"
"        return {}\n"
"    end\n"
"    if type(cfg) ~= 'table' then\n"
"        error(\"Error: cfg should be a table\")\n"
"    end\n"
"    -- just pass {.. dont_check = true, ..} to disable check below\n"
"    if cfg.dont_check then\n"
"        return\n"
"    end\n"
"    local readable_prefix = ''\n"
"    if prefix ~= nil and prefix ~= '' then\n"
"        readable_prefix = prefix .. '.'\n"
"    end\n"
"    local new_cfg = {}\n"
"    for k,v in pairs(cfg) do\n"
"        local readable_name = readable_prefix .. k;\n"
"        if template_cfg[k] == nil then\n"
"            box.error(box.error.CFG, readable_name , \"unexpected option\")\n"
"        elseif v == \"\" or v == nil then\n"
"            -- \"\" and NULL = ffi.cast('void *', 0) set option to default value\n"
"            v = default_cfg[k]\n"
"        elseif template_cfg[k] == 'any' then\n"
"            -- any type is ok\n"
"        elseif type(template_cfg[k]) == 'table' then\n"
"            if type(v) ~= 'table' then\n"
"                box.error(box.error.CFG, readable_name, \"should be a table\")\n"
"            end\n"
"            v = prepare_cfg(v, default_cfg[k], template_cfg[k], modify_cfg[k], readable_name)\n"
"        elseif (string.find(template_cfg[k], ',') == nil) then\n"
"            -- one type\n"
"            if type(v) ~= template_cfg[k] then\n"
"                box.error(box.error.CFG, readable_name, \"should be of type \"..\n"
"                    template_cfg[k])\n"
"            end\n"
"        else\n"
"            local good_types = string.gsub(template_cfg[k], ' ', '');\n"
"            if (string.find(',' .. good_types .. ',', ',' .. type(v) .. ',') == nil) then\n"
"                good_types = string.gsub(good_types, ',', ', ');\n"
"                box.error(box.error.CFG, readable_name, \"should be one of types \"..\n"
"                    template_cfg[k])\n"
"            end\n"
"        end\n"
"        if modify_cfg ~= nil and type(modify_cfg[k]) == 'function' then\n"
"            v = modify_cfg[k](v)\n"
"        end\n"
"        new_cfg[k] = v\n"
"    end\n"
"    return new_cfg\n"
"end\n"
"\n"
"local function apply_default_cfg(cfg, default_cfg)\n"
"    for k,v in pairs(default_cfg) do\n"
"        if cfg[k] == nil then\n"
"            cfg[k] = v\n"
"        elseif type(v) == 'table' then\n"
"            apply_default_cfg(cfg[k], v)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function reload_cfg(oldcfg, cfg)\n"
"    cfg = upgrade_cfg(cfg, translate_cfg)\n"
"    local newcfg = prepare_cfg(cfg, default_cfg, template_cfg, modify_cfg)\n"
"    -- iterate over original table because prepare_cfg() may store NILs\n"
"    for key, val in pairs(cfg) do\n"
"        if dynamic_cfg[key] == nil and oldcfg[key] ~= val then\n"
"            box.error(box.error.RELOAD_CFG, key);\n"
"        end\n"
"    end\n"
"    for key in pairs(cfg) do\n"
"        local val = newcfg[key]\n"
"        local oldval = oldcfg[key]\n"
"        if oldval ~= val then\n"
"            rawset(oldcfg, key, val)\n"
"            if not pcall(dynamic_cfg[key]) then\n"
"                rawset(oldcfg, key, oldval) -- revert the old value\n"
"                return box.error() -- re-throw\n"
"            end\n"
"            if log_cfg_option[key] ~= nil then\n"
"                val = log_cfg_option[key](val)\n"
"            end\n"
"            log.info(\"set '%s' configuration option to %s\", key,\n"
"                json.encode(val))\n"
"        end\n"
"    end\n"
"    if type(box.on_reload_configuration) == 'function' then\n"
"        box.on_reload_configuration()\n"
"    end\n"
"end\n"
"\n"
"local box_cfg_guard_whitelist = {\n"
"    error = true;\n"
"    internal = true;\n"
"    index = true;\n"
"    session = true;\n"
"    tuple = true;\n"
"    runtime = true;\n"
"};\n"
"\n"
"local box = require('box')\n"
"-- Move all box members except 'error' to box_configured\n"
"local box_configured = {}\n"
"for k, v in pairs(box) do\n"
"    box_configured[k] = v\n"
"    if not box_cfg_guard_whitelist[k] then\n"
"        box[k] = nil\n"
"    end\n"
"end\n"
"\n"
"setmetatable(box, {\n"
"    __index = function(table, index)\n"
"        error(debug.traceback(\"Please call box.cfg{} first\"))\n"
"        error(\"Please call box.cfg{} first\")\n"
"     end\n"
"})\n"
"\n"
"local function load_cfg(cfg)\n"
"    cfg = upgrade_cfg(cfg, translate_cfg)\n"
"    cfg = prepare_cfg(cfg, default_cfg, template_cfg, modify_cfg)\n"
"    apply_default_cfg(cfg, default_cfg);\n"
"    -- Save new box.cfg\n"
"    box.cfg = cfg\n"
"    if not pcall(private.cfg_check)  then\n"
"        box.cfg = load_cfg -- restore original box.cfg\n"
"        return box.error() -- re-throw exception from check_cfg()\n"
"    end\n"
"    -- Restore box members after initial configuration\n"
"    for k, v in pairs(box_configured) do\n"
"        box[k] = v\n"
"    end\n"
"    setmetatable(box, nil)\n"
"    box_configured = nil\n"
"    box.cfg = setmetatable(cfg,\n"
"        {\n"
"            __newindex = function(table, index)\n"
"                error('Attempt to modify a read-only table')\n"
"            end,\n"
"            __call = reload_cfg,\n"
"        })\n"
"    private.cfg_load()\n"
"    for key, fun in pairs(dynamic_cfg) do\n"
"        local val = cfg[key]\n"
"        if val ~= nil and not dynamic_cfg_skip_at_load[key] then\n"
"            fun()\n"
"            if val ~= default_cfg[key] then\n"
"                log.info(\"set '%s' configuration option to %s\", key, json.encode(val))\n"
"            end\n"
"        end\n"
"    end\n"
"    if not box.cfg.read_only then\n"
"        box.schema.upgrade{auto = true}\n"
"    end\n"
"end\n"
"box.cfg = load_cfg\n"
"\n"
"-- gh-810:\n"
"-- hack luajit default cpath\n"
"-- commented out because we fixed luajit to build properly, see\n"
"-- https://github.com/luajit/luajit/issues/76\n"
"-- local format = require('tarantool').build.mod_format\n"
"-- package.cpath = package.cpath:gsub(\n"
"--     '?.so', '?.' .. format\n"
"-- ):gsub('loadall.so', 'loadall.' .. format)\n"
""
;
