const char schema_lua[] =
"-- schema.lua (internal file)\n"
"--\n"
"local ffi = require('ffi')\n"
"local msgpack = require('msgpack')\n"
"local msgpackffi = require('msgpackffi')\n"
"local fun = require('fun')\n"
"local log = require('log')\n"
"local session = box.session\n"
"local internal = require('box.internal')\n"
"local function setmap(table)\n"
"    return setmetatable(table, { __serialize = 'map' })\n"
"end\n"
"\n"
"local builtin = ffi.C\n"
"\n"
"-- performance fixup for hot functions\n"
"local tuple_encode = box.tuple.encode\n"
"local tuple_bless = box.tuple.bless\n"
"local is_tuple = box.tuple.is\n"
"assert(tuple_encode ~= nil and tuple_bless ~= nil and is_tuple ~= nil)\n"
"\n"
"ffi.cdef[[\n"
"    struct space *space_by_id(uint32_t id);\n"
"    extern uint32_t box_schema_version();\n"
"    void space_run_triggers(struct space *space, bool yesno);\n"
"    size_t space_bsize(struct space *space);\n"
"\n"
"    typedef struct tuple box_tuple_t;\n"
"    typedef struct iterator box_iterator_t;\n"
"\n"
"    /** \\cond public */\n"
"    box_iterator_t *\n"
"    box_index_iterator(uint32_t space_id, uint32_t index_id, int type,\n"
"                       const char *key, const char *key_end);\n"
"    int\n"
"    box_iterator_next(box_iterator_t *itr, box_tuple_t **result);\n"
"    void\n"
"    box_iterator_free(box_iterator_t *itr);\n"
"    /** \\endcond public */\n"
"    /** \\cond public */\n"
"    ssize_t\n"
"    box_index_len(uint32_t space_id, uint32_t index_id);\n"
"    ssize_t\n"
"    box_index_bsize(uint32_t space_id, uint32_t index_id);\n"
"    int\n"
"    box_index_random(uint32_t space_id, uint32_t index_id, uint32_t rnd,\n"
"                     box_tuple_t **result);\n"
"    int\n"
"    box_index_get(uint32_t space_id, uint32_t index_id, const char *key,\n"
"                  const char *key_end, box_tuple_t **result);\n"
"    int\n"
"    box_index_min(uint32_t space_id, uint32_t index_id, const char *key,\n"
"                  const char *key_end, box_tuple_t **result);\n"
"    int\n"
"    box_index_max(uint32_t space_id, uint32_t index_id, const char *key,\n"
"                  const char *key_end, box_tuple_t **result);\n"
"    ssize_t\n"
"    box_index_count(uint32_t space_id, uint32_t index_id, int type,\n"
"                    const char *key, const char *key_end);\n"
"    /** \\endcond public */\n"
"    /** \\cond public */\n"
"    int\n"
"    box_txn_begin();\n"
"    /** \\endcond public */\n"
"\n"
"    struct port_entry {\n"
"        struct port_entry *next;\n"
"        struct tuple *tuple;\n"
"    };\n"
"\n"
"    struct port {\n"
"        size_t size;\n"
"        struct port_entry *first;\n"
"        struct port_entry *last;\n"
"        struct port_entry first_entry;\n"
"    };\n"
"\n"
"    void\n"
"    port_create(struct port *port);\n"
"\n"
"    void\n"
"    port_destroy(struct port *port);\n"
"\n"
"    int\n"
"    box_select(struct port *port, uint32_t space_id, uint32_t index_id,\n"
"               int iterator, uint32_t offset, uint32_t limit,\n"
"               const char *key, const char *key_end);\n"
"    void password_prepare(const char *password, int len,\n"
"                          char *out, int out_len);\n"
"]]\n"
"\n"
"local function user_or_role_resolve(user)\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    local tuple\n"
"    if type(user) == 'string' then\n"
"        tuple = _user.index.name:get{user}\n"
"    else\n"
"        tuple = _user:get{user}\n"
"    end\n"
"    if tuple == nil then\n"
"        return nil\n"
"    end\n"
"    return tuple[1]\n"
"end\n"
"\n"
"local function role_resolve(name_or_id)\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    local tuple\n"
"    if type(name_or_id) == 'string' then\n"
"        tuple = _user.index.name:get{name_or_id}\n"
"    elseif type(name_or_id) ~= 'nil' then\n"
"        tuple = _user:get{name_or_id}\n"
"    end\n"
"    if tuple == nil or tuple[4] ~= 'role' then\n"
"        return nil\n"
"    else\n"
"        return tuple[1]\n"
"    end\n"
"end\n"
"\n"
"local function user_resolve(name_or_id)\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    local tuple\n"
"    if type(name_or_id) == 'string' then\n"
"        tuple = _user.index.name:get{name_or_id}\n"
"    elseif type(name_or_id) ~= 'nil' then\n"
"        tuple = _user:get{name_or_id}\n"
"    end\n"
"    if tuple == nil or tuple[4] ~= 'user' then\n"
"        return nil\n"
"    else\n"
"        return tuple[1]\n"
"    end\n"
"end\n"
"\n"
"--[[\n"
" @brief Common function to check table with parameters (like options)\n"
" @param table - table with parameters\n"
" @param template - table with expected types of expected parameters\n"
"  type could be comma separated string with lua types (number, string etc),\n"
"  or 'any' if any type is allowed\n"
" The function checks following:\n"
" 1)that parameters table is a table (or nil)\n"
" 2)all keys in parameters are present in template\n"
" 3)type of every parameter fits (one of) types described in template\n"
" Check (2) and (3) could be disabled by adding {, dont_check = <smth is true>, }\n"
"  into parameters table\n"
" The functions calls box.error(box.error.ILLEGAL_PARAMS, ..) on error\n"
" @example\n"
" check_param_table(options, { user = 'string',\n"
"                              port = 'string, number',\n"
"                              data = 'any'} )\n"
"--]]\n"
"local function check_param_table(table, template)\n"
"    if table == nil then\n"
"        return\n"
"    end\n"
"    if type(table) ~= 'table' then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"                  \"options should be a table\")\n"
"    end\n"
"    -- just pass {.. dont_check = true, ..} to disable checks below\n"
"    if table.dont_check then\n"
"        return\n"
"    end\n"
"    for k,v in pairs(table) do\n"
"        if template[k] == nil then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"unexpected option '\" .. k .. \"'\")\n"
"        elseif template[k] == 'any' then\n"
"            -- any type is ok\n"
"        elseif (string.find(template[k], ',') == nil) then\n"
"            -- one type\n"
"            if type(v) ~= template[k] then\n"
"                box.error(box.error.ILLEGAL_PARAMS,\n"
"                          \"options parameter '\" .. k ..\n"
"                          \"' should be of type \" .. template[k])\n"
"            end\n"
"        else\n"
"            local good_types = string.gsub(template[k], ' ', '')\n"
"            local haystack = ',' .. good_types .. ','\n"
"            local needle = ',' .. type(v) .. ','\n"
"            if (string.find(haystack, needle) == nil) then\n"
"                good_types = string.gsub(good_types, ',', ', ')\n"
"                box.error(box.error.ILLEGAL_PARAMS,\n"
"                          \"options parameter '\" .. k ..\n"
"                          \"' should be one of types: \" .. template[k])\n"
"            end\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"--[[\n"
" @brief Common function to check type parameter (of function)\n"
" Calls box.error(box.error.ILLEGAL_PARAMS, ) on error\n"
" @example: check_param(user, 'user', 'string')\n"
"--]]\n"
"local function check_param(param, name, should_be_type)\n"
"    if type(param) ~= should_be_type then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"                  name .. \" should be a \" .. should_be_type)\n"
"    end\n"
"end\n"
"\n"
"--[[\n"
" Adds to a table key-value pairs from defaults table\n"
"  that is not present in original table.\n"
" Returns updated table.\n"
" If nil is passed instead of table, it's treated as empty table {}\n"
" For example update_param_table({ type = 'hash', temporary = true },\n"
"                                { type = 'tree', unique = true })\n"
"  will return table { type = 'hash', temporary = true, unique = true }\n"
"--]]\n"
"local function update_param_table(table, defaults)\n"
"    local new_table = {}\n"
"    if defaults ~= nil then\n"
"        for k,v in pairs(defaults) do\n"
"            new_table[k] = v\n"
"        end\n"
"    end\n"
"    if table ~= nil then\n"
"        for k,v in pairs(table) do\n"
"            new_table[k] = v\n"
"        end\n"
"    end\n"
"    return new_table\n"
"end\n"
"\n"
"box.begin = function()\n"
"    if builtin.box_txn_begin() == -1 then\n"
"        box.error()\n"
"    end\n"
"end\n"
"\n"
"local function atomic_tail(status, ...)\n"
"    if not status then\n"
"        box.rollback()\n"
"        error((...), 2)\n"
"     end\n"
"     box.commit()\n"
"     return ...\n"
"end\n"
"\n"
"box.atomic = function(fun, ...)\n"
"    box.begin()\n"
"    return atomic_tail(pcall(fun, ...))\n"
"end\n"
"\n"
"-- box.commit yields, so it's defined as Lua/C binding\n"
"-- box.rollback yields as well\n"
"\n"
"box.schema.space = {}\n"
"box.schema.space.create = function(name, options)\n"
"    check_param(name, 'name', 'string')\n"
"    local options_template = {\n"
"        if_not_exists = 'boolean',\n"
"        engine = 'string',\n"
"        id = 'number',\n"
"        field_count = 'number',\n"
"        user = 'string, number',\n"
"        format = 'table',\n"
"        temporary = 'boolean',\n"
"    }\n"
"    local options_defaults = {\n"
"        engine = 'memtx',\n"
"        field_count = 0,\n"
"        temporary = false,\n"
"    }\n"
"    check_param_table(options, options_template)\n"
"    options = update_param_table(options, options_defaults)\n"
"\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    if box.space[name] then\n"
"        if options.if_not_exists then\n"
"            return box.space[name], \"not created\"\n"
"        else\n"
"            box.error(box.error.SPACE_EXISTS, name)\n"
"        end\n"
"    end\n"
"    local id = options.id\n"
"    if not id then\n"
"        local _schema = box.space._schema\n"
"        local max_id = _schema:update({'max_id'}, {{'+', 2, 1}})\n"
"        if max_id == nil then\n"
"            id = _space.index.primary:max()[1]\n"
"            if id < box.schema.SYSTEM_ID_MAX then\n"
"                id = box.schema.SYSTEM_ID_MAX\n"
"            end\n"
"            max_id = _schema:insert{'max_id', id + 1}\n"
"        end\n"
"        id = max_id[2]\n"
"    end\n"
"    local uid = nil\n"
"    if options.user then\n"
"        uid = user_or_role_resolve(options.user)\n"
"    end\n"
"    if uid == nil then\n"
"        uid = session.uid()\n"
"    end\n"
"    local format = options.format and options.format or {}\n"
"    -- filter out global parameters from the options array\n"
"    local space_options = setmap({\n"
"        temporary = options.temporary and true or nil,\n"
"    })\n"
"    _space:insert{id, uid, name, options.engine, options.field_count,\n"
"        space_options, format}\n"
"    return box.space[id], \"created\"\n"
"end\n"
"\n"
"-- space format - the metadata about space fields\n"
"function box.schema.space.format(id, format)\n"
"    local _space = box.space._space\n"
"    check_param(id, 'id', 'number')\n"
"\n"
"    if format == nil then\n"
"        return _space:get(id)[7]\n"
"    else\n"
"        check_param(format, 'format', 'table')\n"
"        _space:update(id, {{'=', 7, format}})\n"
"    end\n"
"end\n"
"\n"
"box.schema.create_space = box.schema.space.create\n"
"\n"
"box.schema.space.drop = function(space_id, space_name, opts)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_exists = 'boolean' })\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local _truncate = box.space[box.schema.TRUNCATE_ID]\n"
"    local keys = _index:select(space_id)\n"
"    for i = #keys, 1, -1 do\n"
"        local v = keys[i]\n"
"        _index:delete{v[1], v[2]}\n"
"    end\n"
"    local privs = _priv.index.object:select{'space', space_id}\n"
"    for k, tuple in pairs(privs) do\n"
"        box.schema.user.revoke(tuple[2], tuple[5], tuple[3], tuple[4])\n"
"    end\n"
"    if _space:delete{space_id} == nil then\n"
"        if space_name == nil then\n"
"            space_name = '#'..tostring(space_id)\n"
"        end\n"
"        if not opts.if_exists then\n"
"            box.error(box.error.NO_SUCH_SPACE, space_name)\n"
"        end\n"
"    end\n"
"    _truncate:delete{space_id}\n"
"end\n"
"\n"
"box.schema.space.rename = function(space_id, space_name)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    check_param(space_name, 'space_name', 'string')\n"
"\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    _space:update(space_id, {{\"=\", 3, space_name}})\n"
"end\n"
"\n"
"box.schema.index = {}\n"
"\n"
"local function check_index_parts(parts)\n"
"    if type(parts) ~= \"table\" then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"                  \"options.parts parameter should be a table\")\n"
"    end\n"
"    if #parts % 2 ~= 0 then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"                  \"options.parts: expected field_no (number), type (string) pairs\")\n"
"    end\n"
"    for i=1,#parts,2 do\n"
"        if type(parts[i]) ~= \"number\" then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"options.parts: expected field_no (number), type (string) pairs\")\n"
"        elseif parts[i] == 0 then\n"
"            -- Lua uses one-based field numbers but _space is zero-based\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"invalid index parts: field_no must be one-based\")\n"
"        end\n"
"    end\n"
"    for i=2,#parts,2 do\n"
"        if type(parts[i]) ~= \"string\" then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"options.parts: expected field_no (number), type (string) pairs\")\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function update_index_parts(parts)\n"
"    local new_parts = {}\n"
"    for i=1,#parts do\n"
"        -- Lua uses one-based field numbers but _space is zero-based\n"
"        if i % 2 == 1 then\n"
"            new_parts[i] = parts[i] - 1\n"
"        else\n"
"            new_parts[i] = parts[i]\n"
"        end\n"
"    end\n"
"    return new_parts\n"
"end\n"
"\n"
"-- Historically, some properties of an index\n"
"-- are stored as tuple fields, others in a\n"
"-- single field containing msgpack map.\n"
"-- This is the map.\n"
"local index_options = {\n"
"    unique = 'boolean',\n"
"    dimension = 'number',\n"
"    distance = 'string',\n"
"    run_count_per_level = 'number',\n"
"    run_size_ratio = 'number',\n"
"    range_size = 'number',\n"
"    page_size = 'number',\n"
"    bloom_fpr = 'number',\n"
"}\n"
"\n"
"--\n"
"-- check_param_table() template for alter index,\n"
"-- includes all index options.\n"
"--\n"
"local alter_index_template = {\n"
"    id = 'number',\n"
"    name = 'string',\n"
"    type = 'string',\n"
"    parts = 'table',\n"
"}\n"
"for k, v in pairs(index_options) do\n"
"    alter_index_template[k] = v\n"
"end\n"
"\n"
"--\n"
"-- check_param_table() template for create_index(), includes\n"
"-- all index options and if_not_exists specifier\n"
"--\n"
"local create_index_template = table.deepcopy(alter_index_template)\n"
"create_index_template.if_not_exists = \"boolean\"\n"
"\n"
"box.schema.index.create = function(space_id, name, options)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    check_param(name, 'name', 'string')\n"
"    check_param_table(options, create_index_template)\n"
"\n"
"    local options_defaults = {\n"
"        type = 'tree',\n"
"    }\n"
"    options = update_param_table(options, options_defaults)\n"
"    local type_dependent_defaults = {\n"
"        rtree = {parts = { 2, 'array' }, unique = false},\n"
"        bitset = {parts = { 2, 'unsigned' }, unique = false},\n"
"        other = {parts = { 1, 'unsigned' }, unique = true},\n"
"    }\n"
"    options_defaults = type_dependent_defaults[options.type]\n"
"            and type_dependent_defaults[options.type]\n"
"            or type_dependent_defaults.other\n"
"    options = update_param_table(options, options_defaults)\n"
"    if box.space[space_id] ~= nil and box.space[space_id].engine == 'vinyl' then\n"
"        options_defaults = {\n"
"            page_size = box.cfg.vinyl_page_size,\n"
"            range_size = box.cfg.vinyl_range_size,\n"
"            run_count_per_level = box.cfg.vinyl_run_count_per_level,\n"
"            run_size_ratio = box.cfg.vinyl_run_size_ratio,\n"
"            bloom_fpr = box.cfg.vinyl_bloom_fpr\n"
"        }\n"
"    else\n"
"        options_defaults = {}\n"
"    end\n"
"    options = update_param_table(options, options_defaults)\n"
"\n"
"    check_index_parts(options.parts)\n"
"    options.parts = update_index_parts(options.parts)\n"
"\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    if _index.index.name:get{space_id, name} then\n"
"        if options.if_not_exists then\n"
"            return box.space[space_id].index[name], \"not created\"\n"
"        else\n"
"            box.error(box.error.INDEX_EXISTS, name)\n"
"        end\n"
"    end\n"
"\n"
"    local iid = 0\n"
"    if options.id then\n"
"        iid = options.id\n"
"    else\n"
"        -- max\n"
"        local tuple = _index.index[0]\n"
"            :select(space_id, { limit = 1, iterator = 'LE' })[1]\n"
"        if tuple then\n"
"            local id = tuple[1]\n"
"            if id == space_id then\n"
"                iid = tuple[2] + 1\n"
"            end\n"
"        end\n"
"    end\n"
"    local parts = {}\n"
"    for i = 1, #options.parts, 2 do\n"
"        table.insert(parts, {options.parts[i], options.parts[i + 1]})\n"
"    end\n"
"    -- create_index() options contains type, parts, etc,\n"
"    -- stored separately. Remove these members from index_opts\n"
"    local index_opts = {\n"
"            dimension = options.dimension,\n"
"            unique = options.unique,\n"
"            distance = options.distance,\n"
"            page_size = options.page_size,\n"
"            range_size = options.range_size,\n"
"            run_count_per_level = options.run_count_per_level,\n"
"            run_size_ratio = options.run_size_ratio,\n"
"            bloom_fpr = options.bloom_fpr,\n"
"    }\n"
"    local field_type_aliases = {\n"
"        num = 'unsigned'; -- Deprecated since 1.7.2\n"
"        uint = 'unsigned';\n"
"        str = 'string';\n"
"        int = 'integer';\n"
"    };\n"
"    for _, part in pairs(parts) do\n"
"        local field_type = part[2]:lower()\n"
"        part[2] = field_type_aliases[field_type] or field_type\n"
"        if field_type == 'num' then\n"
"            log.warn(\"field type '%s' is deprecated since Tarantool 1.7, \"..\n"
"                     \"please use '%s' instead\", field_type, part[2])\n"
"        end\n"
"    end\n"
"    _index:insert{space_id, iid, name, options.type, index_opts, parts}\n"
"    return box.space[space_id].index[name]\n"
"end\n"
"\n"
"box.schema.index.drop = function(space_id, index_id)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    check_param(index_id, 'index_id', 'number')\n"
"\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    _index:delete{space_id, index_id}\n"
"end\n"
"\n"
"box.schema.index.rename = function(space_id, index_id, name)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    check_param(index_id, 'index_id', 'number')\n"
"    check_param(name, 'name', 'string')\n"
"\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    _index:update({space_id, index_id}, {{\"=\", 3, name}})\n"
"end\n"
"\n"
"box.schema.index.alter = function(space_id, index_id, options)\n"
"    if box.space[space_id] == nil then\n"
"        box.error(box.error.NO_SUCH_SPACE, '#'..tostring(space_id))\n"
"    end\n"
"    if box.space[space_id].index[index_id] == nil then\n"
"        box.error(box.error.NO_SUCH_INDEX, index_id, box.space[space_id].name)\n"
"    end\n"
"    if options == nil then\n"
"        return\n"
"    end\n"
"\n"
"    check_param_table(options, alter_index_template)\n"
"\n"
"    if type(space_id) ~= \"number\" then\n"
"        space_id = box.space[space_id].id\n"
"    end\n"
"    if type(index_id) ~= \"number\" then\n"
"        index_id = box.space[space_id].index[index_id].id\n"
"    end\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    if options.id ~= nil then\n"
"        local can_update_field = {id = true, name = true, type = true }\n"
"        local can_update = true\n"
"        local cant_update_fields = ''\n"
"        for k,v in pairs(options) do\n"
"            if not can_update_field[k] then\n"
"                can_update = false\n"
"                cant_update_fields = cant_update_fields .. ' ' .. k\n"
"            end\n"
"        end\n"
"        if not can_update then\n"
"            box.error(box.error.PROC_LUA,\n"
"                      \"Don't know how to update both id and\" ..\n"
"                       cant_update_fields)\n"
"        end\n"
"        local ops = {}\n"
"        local function add_op(value, field_no)\n"
"            if value then\n"
"                table.insert(ops, {'=', field_no, value})\n"
"            end\n"
"        end\n"
"        add_op(options.id, 2)\n"
"        add_op(options.name, 3)\n"
"        add_op(options.type, 4)\n"
"        _index:update({space_id, index_id}, ops)\n"
"        return\n"
"    end\n"
"    local tuple = _index:get{space_id, index_id }\n"
"    local parts = {}\n"
"    local index_opts = {}\n"
"    local OPTS = 5\n"
"    local PARTS = 6\n"
"    if type(tuple[OPTS]) == 'number' then\n"
"        -- old format\n"
"        index_opts.unique = tuple[OPTS] == 1\n"
"        local part_count = tuple[PARTS]\n"
"        for i = 1, part_count do\n"
"            table.insert(parts, {tuple[2 * i + 4], tuple[2 * i + 5]});\n"
"        end\n"
"    else\n"
"        -- new format\n"
"        index_opts = tuple[OPTS]\n"
"        parts = tuple[PARTS]\n"
"    end\n"
"    if options.name == nil then\n"
"        options.name = tuple[3]\n"
"    end\n"
"    if options.type == nil then\n"
"        options.type = tuple[4]\n"
"    end\n"
"    for k, t in pairs(index_options) do\n"
"        if options[k] ~= nil then\n"
"            index_opts[k] = options[k]\n"
"        end\n"
"    end\n"
"    if options.parts ~= nil then\n"
"        check_index_parts(options.parts)\n"
"        options.parts = update_index_parts(options.parts)\n"
"        parts = {}\n"
"        for i = 1, #options.parts, 2 do\n"
"            table.insert(parts, {options.parts[i], options.parts[i + 1]})\n"
"        end\n"
"    end\n"
"    _index:replace{space_id, index_id, options.name, options.type,\n"
"                   index_opts, parts}\n"
"end\n"
"\n"
"-- a static box_tuple_t ** instance for calling box_index_* API\n"
"local ptuple = ffi.new('box_tuple_t *[1]')\n"
"\n"
"local function keify(key)\n"
"    if key == nil then\n"
"        return {}\n"
"    elseif type(key) == \"table\" or is_tuple(key) then\n"
"        return key\n"
"    end\n"
"    return {key}\n"
"end\n"
"\n"
"local iterator_t = ffi.typeof('struct iterator')\n"
"ffi.metatype(iterator_t, {\n"
"    __tostring = function(iterator)\n"
"        return \"<iterator state>\"\n"
"    end;\n"
"})\n"
"\n"
"local iterator_gen = function(param, state)\n"
"    --[[\n"
"        index:pairs() mostly conforms to the Lua for-in loop conventions and\n"
"        tries to follow the best practices of Lua community.\n"
"\n"
"        - this generating function is stateless.\n"
"\n"
"        - *param* should contain **immutable** data needed to fully define\n"
"          an iterator. *param* is opaque for users. Currently it contains keybuf\n"
"          string just to prevent GC from collecting it. In future some other\n"
"          variables like space_id, index_id, sc_version will be stored here.\n"
"\n"
"        - *state* should contain **immutable** transient state of an iterator.\n"
"          *state* is opaque for users. Currently it contains `struct iterator`\n"
"          cdata that is modified during iteration. This is a sad limitation of\n"
"          underlying C API. Moreover, the separation of *param* and *state* is\n"
"          not properly implemented here. These drawbacks can be fixed in\n"
"          future without changing this API.\n"
"\n"
"        Please check out http://www.lua.org/pil/7.3.html for details.\n"
"    --]]\n"
"    if not ffi.istype(iterator_t, state) then\n"
"        error('usage: next(param, state)')\n"
"    end\n"
"    -- next() modifies state in-place\n"
"    if builtin.box_iterator_next(state, ptuple) ~= 0 then\n"
"        return box.error() -- error\n"
"    elseif ptuple[0] ~= nil then\n"
"        return state, tuple_bless(ptuple[0]) -- new state, value\n"
"    else\n"
"        return nil\n"
"    end\n"
"end\n"
"\n"
"local iterator_gen_luac = function(param, state)\n"
"    local tuple = internal.iterator_next(state)\n"
"    if tuple ~= nil then\n"
"        return state, tuple -- new state, value\n"
"    else\n"
"        return nil\n"
"    end\n"
"end\n"
"\n"
"-- global struct port instance to use by select()/get()\n"
"local port = ffi.new('struct port')\n"
"local port_entry_t = ffi.typeof('struct port_entry')\n"
"\n"
"-- Helper function to check space:method() usage\n"
"local function check_space_arg(space, method)\n"
"    if type(space) ~= 'table' or space.id == nil then\n"
"        local fmt = 'Use space:%s(...) instead of space.%s(...)'\n"
"        error(string.format(fmt, method, method))\n"
"    end\n"
"end\n"
"box.internal.check_space_arg = check_space_arg -- for net.box\n"
"\n"
"-- Helper function for nicer error messages\n"
"-- in some cases when space object is misused\n"
"-- Takes time so should not be used for DML.\n"
"local function check_space_exists(space)\n"
"    local s = box.space[space.id]\n"
"    if s == nil then\n"
"        box.error(box.error.NO_SUCH_SPACE, space.name)\n"
"    end\n"
"end\n"
"\n"
"-- Helper function to check index:method() usage\n"
"local function check_index_arg(index, method)\n"
"    if type(index) ~= 'table' or index.id == nil then\n"
"        local fmt = 'Use index:%s(...) instead of index.%s(...)'\n"
"        error(string.format(fmt, method, method))\n"
"    end\n"
"end\n"
"box.internal.check_index_arg = check_index_arg -- for net.box\n"
"\n"
"-- Helper function to check that space have primary key and return it\n"
"local function check_primary_index(space)\n"
"    local pk = space.index[0]\n"
"    if pk == nil then\n"
"        box.error(box.error.NO_SUCH_INDEX, 0, space.name)\n"
"    end\n"
"    return pk\n"
"end\n"
"box.internal.check_primary_index = check_primary_index -- for net.box\n"
"\n"
"box.internal.schema_version = builtin.box_schema_version\n"
"\n"
"local function check_iterator_type(opts, key_is_nil)\n"
"    local itype\n"
"    if opts and opts.iterator then\n"
"        if type(opts.iterator) == \"number\" then\n"
"            itype = opts.iterator\n"
"        elseif type(opts.iterator) == \"string\" then\n"
"            itype = box.index[string.upper(opts.iterator)]\n"
"            if itype == nil then\n"
"                box.error(box.error.ITERATOR_TYPE, opts.iterator)\n"
"            end\n"
"        else\n"
"            box.error(box.error.ITERATOR_TYPE, tostring(opts.iterator))\n"
"        end\n"
"    elseif opts and type(opts) == \"string\" then\n"
"        itype = box.index[string.upper(opts)]\n"
"        if itype == nil then\n"
"            box.error(box.error.ITERATOR_TYPE, opts)\n"
"        end\n"
"    else\n"
"        -- Use ALL for {} and nil keys and EQ for other keys\n"
"        itype = key_is_nil and box.index.ALL or box.index.EQ\n"
"    end\n"
"    return itype\n"
"end\n"
"\n"
"internal.check_iterator_type = check_iterator_type -- export for net.box\n"
"\n"
"function box.schema.space.bless(space)\n"
"    local index_mt = {}\n"
"    -- __len and __index\n"
"    index_mt.len = function(index)\n"
"        check_index_arg(index, 'len')\n"
"        local ret = builtin.box_index_len(index.space_id, index.id)\n"
"        if ret == -1 then\n"
"            box.error()\n"
"        end\n"
"        return tonumber(ret)\n"
"    end\n"
"    -- index.bsize\n"
"    index_mt.bsize = function(index)\n"
"        check_index_arg(index, 'bsize')\n"
"        local ret = builtin.box_index_bsize(index.space_id, index.id)\n"
"        if ret == -1 then\n"
"            box.error()\n"
"        end\n"
"        return tonumber(ret)\n"
"    end\n"
"    index_mt.__len = index_mt.len -- Lua 5.2 compatibility\n"
"    index_mt.__newindex = function(table, index)\n"
"        return error('Attempt to modify a read-only table') end\n"
"    index_mt.__index = index_mt\n"
"    -- min and max\n"
"    index_mt.min_ffi = function(index, key)\n"
"        check_index_arg(index, 'min')\n"
"        local pkey, pkey_end = tuple_encode(key)\n"
"        if builtin.box_index_min(index.space_id, index.id,\n"
"                                 pkey, pkey_end, ptuple) ~= 0 then\n"
"            box.error() -- error\n"
"        elseif ptuple[0] ~= nil then\n"
"            return tuple_bless(ptuple[0])\n"
"        else\n"
"            return\n"
"        end\n"
"    end\n"
"    index_mt.min_luac = function(index, key)\n"
"        check_index_arg(index, 'min')\n"
"        key = keify(key)\n"
"        return internal.min(index.space_id, index.id, key);\n"
"    end\n"
"    index_mt.max_ffi = function(index, key)\n"
"        check_index_arg(index, 'max')\n"
"        local pkey, pkey_end = tuple_encode(key)\n"
"        if builtin.box_index_max(index.space_id, index.id,\n"
"                                 pkey, pkey_end, ptuple) ~= 0 then\n"
"            box.error() -- error\n"
"        elseif ptuple[0] ~= nil then\n"
"            return tuple_bless(ptuple[0])\n"
"        else\n"
"            return\n"
"        end\n"
"    end\n"
"    index_mt.max_luac = function(index, key)\n"
"        check_index_arg(index, 'max')\n"
"        key = keify(key)\n"
"        return internal.max(index.space_id, index.id, key);\n"
"    end\n"
"    index_mt.random_ffi = function(index, rnd)\n"
"        check_index_arg(index, 'random')\n"
"        rnd = rnd or math.random()\n"
"        if builtin.box_index_random(index.space_id, index.id, rnd,\n"
"                                    ptuple) ~= 0 then\n"
"            box.error() -- error\n"
"        elseif ptuple[0] ~= nil then\n"
"            return tuple_bless(ptuple[0])\n"
"        else\n"
"            return\n"
"        end\n"
"    end\n"
"    index_mt.random_luac = function(index, rnd)\n"
"        check_index_arg(index, 'random')\n"
"        rnd = rnd or math.random()\n"
"        return internal.random(index.space_id, index.id, rnd);\n"
"    end\n"
"    -- iteration\n"
"    index_mt.pairs_ffi = function(index, key, opts)\n"
"        check_index_arg(index, 'pairs')\n"
"        local pkey, pkey_end = tuple_encode(key)\n"
"        local itype = check_iterator_type(opts, pkey + 1 >= pkey_end);\n"
"\n"
"        local keybuf = ffi.string(pkey, pkey_end - pkey)\n"
"        local pkeybuf = ffi.cast('const char *', keybuf)\n"
"        local cdata = builtin.box_index_iterator(index.space_id, index.id,\n"
"            itype, pkeybuf, pkeybuf + #keybuf);\n"
"        if cdata == nil then\n"
"            box.error()\n"
"        end\n"
"        return fun.wrap(iterator_gen, keybuf,\n"
"            ffi.gc(cdata, builtin.box_iterator_free))\n"
"    end\n"
"    index_mt.pairs_luac = function(index, key, opts)\n"
"        check_index_arg(index, 'pairs')\n"
"        key = keify(key)\n"
"        local itype = check_iterator_type(opts, #key == 0);\n"
"        local keymp = msgpack.encode(key)\n"
"        local keybuf = ffi.string(keymp, #keymp)\n"
"        local cdata = internal.iterator(index.space_id, index.id, itype, keymp);\n"
"        return fun.wrap(iterator_gen_luac, keybuf,\n"
"            ffi.gc(cdata, builtin.box_iterator_free))\n"
"    end\n"
"\n"
"    -- index subtree size\n"
"    index_mt.count_ffi = function(index, key, opts)\n"
"        check_index_arg(index, 'count')\n"
"        local pkey, pkey_end = tuple_encode(key)\n"
"        local itype = check_iterator_type(opts, pkey + 1 >= pkey_end);\n"
"        local count = builtin.box_index_count(index.space_id, index.id,\n"
"            itype, pkey, pkey_end);\n"
"        if count == -1 then\n"
"            box.error()\n"
"        end\n"
"        return tonumber(count)\n"
"    end\n"
"    index_mt.count_luac = function(index, key, opts)\n"
"        check_index_arg(index, 'count')\n"
"        key = keify(key)\n"
"        local itype = check_iterator_type(opts, #key == 0);\n"
"        return internal.count(index.space_id, index.id, itype, key);\n"
"    end\n"
"\n"
"    index_mt.get_ffi = function(index, key)\n"
"        check_index_arg(index, 'get')\n"
"        local key, key_end = tuple_encode(key)\n"
"        if builtin.box_index_get(index.space_id, index.id,\n"
"                                 key, key_end, ptuple) ~= 0 then\n"
"            return box.error() -- error\n"
"        elseif ptuple[0] ~= nil then\n"
"            return tuple_bless(ptuple[0])\n"
"        else\n"
"            return\n"
"        end\n"
"    end\n"
"    index_mt.get_luac = function(index, key)\n"
"        check_index_arg(index, 'get')\n"
"        key = keify(key)\n"
"        return internal.get(index.space_id, index.id, key)\n"
"    end\n"
"\n"
"    local function check_select_opts(opts, key_is_nil)\n"
"        local offset = 0\n"
"        local limit = 4294967295\n"
"        local iterator = check_iterator_type(opts, key_is_nil)\n"
"        if opts ~= nil then\n"
"            if opts.offset ~= nil then\n"
"                offset = opts.offset\n"
"            end\n"
"            if opts.limit ~= nil then\n"
"                limit = opts.limit\n"
"            end\n"
"        end\n"
"        return iterator, offset, limit\n"
"    end\n"
"\n"
"    index_mt.select_ffi = function(index, key, opts)\n"
"        check_index_arg(index, 'select')\n"
"        local key, key_end = tuple_encode(key)\n"
"        local iterator, offset, limit = check_select_opts(opts, key + 1 >= key_end)\n"
"\n"
"        builtin.port_create(port)\n"
"        if builtin.box_select(port, index.space_id,\n"
"            index.id, iterator, offset, limit, key, key_end) ~=0 then\n"
"            builtin.port_destroy(port);\n"
"            return box.error()\n"
"        end\n"
"\n"
"        local ret = {}\n"
"        local entry = port.first\n"
"        for i=1,tonumber(port.size),1 do\n"
"            ret[i] = tuple_bless(entry.tuple)\n"
"            entry = entry.next\n"
"        end\n"
"        builtin.port_destroy(port);\n"
"        return ret\n"
"    end\n"
"\n"
"    index_mt.select_luac = function(index, key, opts)\n"
"        check_index_arg(index, 'select')\n"
"        local key = keify(key)\n"
"        local iterator, offset, limit = check_select_opts(opts, #key == 0)\n"
"        return internal.select(index.space_id, index.id, iterator,\n"
"            offset, limit, key)\n"
"    end\n"
"\n"
"    index_mt.update = function(index, key, ops)\n"
"        check_index_arg(index, 'update')\n"
"        return internal.update(index.space_id, index.id, keify(key), ops);\n"
"    end\n"
"    index_mt.delete = function(index, key)\n"
"        check_index_arg(index, 'delete')\n"
"        return internal.delete(index.space_id, index.id, keify(key));\n"
"    end\n"
"\n"
"    index_mt.info = function(index)\n"
"        return internal.info(index.space_id, index.id);\n"
"    end\n"
"\n"
"    index_mt.drop = function(index)\n"
"        check_index_arg(index, 'drop')\n"
"        return box.schema.index.drop(index.space_id, index.id)\n"
"    end\n"
"    index_mt.rename = function(index, name)\n"
"        check_index_arg(index, 'rename')\n"
"        return box.schema.index.rename(index.space_id, index.id, name)\n"
"    end\n"
"    index_mt.alter = function(index, options)\n"
"        check_index_arg(index, 'alter')\n"
"        if index.id == nil or index.space_id == nil then\n"
"            box.error(box.error.PROC_LUA, \"Usage: index:alter{opts}\")\n"
"        end\n"
"        return box.schema.index.alter(index.space_id, index.id, options)\n"
"    end\n"
"\n"
"    -- true if reading operations may yield\n"
"    local read_yields = space.engine == 'vinyl'\n"
"    local read_ops = {'select', 'get', 'min', 'max', 'count', 'random', 'pairs'}\n"
"    for _, op in ipairs(read_ops) do\n"
"        if read_yields then\n"
"            -- use Lua/C implmenetation\n"
"            index_mt[op] = index_mt[op .. \"_luac\"]\n"
"        else\n"
"            -- use FFI implementation\n"
"            index_mt[op] = index_mt[op .. \"_ffi\"]\n"
"        end\n"
"    end\n"
"    index_mt.__pairs = index_mt.pairs -- Lua 5.2 compatibility\n"
"    index_mt.__ipairs = index_mt.pairs -- Lua 5.2 compatibility\n"
"    --\n"
"    local space_mt = {}\n"
"    space_mt.len = function(space)\n"
"        check_space_arg(space, 'len')\n"
"        local pk = space.index[0]\n"
"        if pk == nil then\n"
"            return 0 -- empty space without indexes, return 0\n"
"        end\n"
"        return space.index[0]:len()\n"
"    end\n"
"    space_mt.count = function(space, key, opts)\n"
"        check_space_arg(space, 'count')\n"
"        local pk = space.index[0]\n"
"        if pk == nil then\n"
"            return 0 -- empty space without indexes, return 0\n"
"        end\n"
"        return pk:count(key, opts)\n"
"    end\n"
"    space_mt.bsize = function(space)\n"
"        check_space_arg(space, 'bsize')\n"
"        local s = builtin.space_by_id(space.id)\n"
"        if s == nil then\n"
"            box.error(box.error.NO_SUCH_SPACE, space.name)\n"
"        end\n"
"        return builtin.space_bsize(s)\n"
"    end\n"
"    space_mt.__newindex = index_mt.__newindex\n"
"\n"
"    space_mt.get = function(space, key)\n"
"        check_space_arg(space, 'get')\n"
"        return check_primary_index(space):get(key)\n"
"    end\n"
"    space_mt.select = function(space, key, opts)\n"
"        check_space_arg(space, 'select')\n"
"        return check_primary_index(space):select(key, opts)\n"
"    end\n"
"    space_mt.insert = function(space, tuple)\n"
"        check_space_arg(space, 'insert')\n"
"        return internal.insert(space.id, tuple);\n"
"    end\n"
"    space_mt.replace = function(space, tuple)\n"
"        check_space_arg(space, 'replace')\n"
"        return internal.replace(space.id, tuple);\n"
"    end\n"
"    space_mt.put = space_mt.replace; -- put is an alias for replace\n"
"    space_mt.update = function(space, key, ops)\n"
"        check_space_arg(space, 'update')\n"
"        return check_primary_index(space):update(key, ops)\n"
"    end\n"
"    space_mt.upsert = function(space, tuple_key, ops, deprecated)\n"
"        check_space_arg(space, 'upsert')\n"
"        if deprecated ~= nil then\n"
"            local msg = \"Error: extra argument in upsert call: \"\n"
"            msg = msg .. tostring(deprecated)\n"
"            msg = msg .. \". Usage :upsert(tuple, operations)\"\n"
"            box.error(box.error.PROC_LUA, msg)\n"
"        end\n"
"        return internal.upsert(space.id, tuple_key, ops);\n"
"    end\n"
"    space_mt.delete = function(space, key)\n"
"        check_space_arg(space, 'delete')\n"
"        return check_primary_index(space):delete(key)\n"
"    end\n"
"-- Assumes that spaceno has a TREE (NUM) primary key\n"
"-- inserts a tuple after getting the next value of the\n"
"-- primary key and returns it back to the user\n"
"    space_mt.auto_increment = function(space, tuple)\n"
"        check_space_arg(space, 'auto_increment')\n"
"        local max_tuple = check_primary_index(space):max()\n"
"        local max = 0\n"
"        if max_tuple ~= nil then\n"
"            max = max_tuple[1]\n"
"        end\n"
"        table.insert(tuple, 1, max + 1)\n"
"        return space:insert(tuple)\n"
"    end\n"
"\n"
"    space_mt.pairs = function(space, key, opts)\n"
"        check_space_arg(space, 'pairs')\n"
"        local pk = space.index[0]\n"
"        if pk == nil then\n"
"            -- empty space without indexes, return empty iterator\n"
"            return fun.iter({})\n"
"        end\n"
"        return pk:pairs(key, opts)\n"
"    end\n"
"    space_mt.__pairs = space_mt.pairs -- Lua 5.2 compatibility\n"
"    space_mt.__ipairs = space_mt.pairs -- Lua 5.2 compatibility\n"
"    space_mt.truncate = function(space)\n"
"        check_space_arg(space, 'truncate')\n"
"        return internal.truncate(space.id)\n"
"    end\n"
"    space_mt.format = function(space, format)\n"
"        check_space_arg(space, 'format')\n"
"        return box.schema.space.format(space.id, format)\n"
"    end\n"
"    space_mt.drop = function(space)\n"
"        check_space_arg(space, 'drop')\n"
"        check_space_exists(space)\n"
"        return box.schema.space.drop(space.id, space.name)\n"
"    end\n"
"    space_mt.rename = function(space, name)\n"
"        check_space_arg(space, 'rename')\n"
"        check_space_exists(space)\n"
"        return box.schema.space.rename(space.id, name)\n"
"    end\n"
"    space_mt.create_index = function(space, name, options)\n"
"        check_space_arg(space, 'create_index')\n"
"        check_space_exists(space)\n"
"        return box.schema.index.create(space.id, name, options)\n"
"    end\n"
"    space_mt.run_triggers = function(space, yesno)\n"
"        check_space_arg(space, 'run_triggers')\n"
"        local s = builtin.space_by_id(space.id)\n"
"        if s == nil then\n"
"            box.error(box.error.NO_SUCH_SPACE, space.name)\n"
"        end\n"
"        builtin.space_run_triggers(s, yesno)\n"
"    end\n"
"    space_mt.__index = space_mt\n"
"\n"
"    setmetatable(space, space_mt)\n"
"    if type(space.index) == 'table' and space.enabled then\n"
"        for j, index in pairs(space.index) do\n"
"            if type(j) == 'number' then\n"
"                setmetatable(index, index_mt)\n"
"            end\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function privilege_resolve(privilege)\n"
"    local numeric = 0\n"
"    if type(privilege) == 'string' then\n"
"        privilege = string.lower(privilege)\n"
"        if string.find(privilege, 'read') then\n"
"            numeric = numeric + 1\n"
"        end\n"
"        if string.find(privilege, 'write') then\n"
"            numeric = numeric + 2\n"
"        end\n"
"        if string.find(privilege, 'execute') then\n"
"            numeric = numeric + 4\n"
"        end\n"
"    else\n"
"        numeric = privilege\n"
"    end\n"
"    return numeric\n"
"end\n"
"\n"
"local function checked_privilege(privilege, object_type)\n"
"    local priv_hex = privilege_resolve(privilege)\n"
"    if object_type == 'role' and priv_hex ~= 4 then\n"
"        box.error(box.error.UNSUPPORTED_ROLE_PRIV, privilege)\n"
"    end\n"
"    return priv_hex\n"
"end\n"
"\n"
"local function privilege_name(privilege)\n"
"    local names = {}\n"
"    if bit.band(privilege, 1) ~= 0 then\n"
"        table.insert(names, \"read\")\n"
"    end\n"
"    if bit.band(privilege, 2) ~= 0 then\n"
"        table.insert(names, \"write\")\n"
"    end\n"
"    if bit.band(privilege, 4) ~= 0 then\n"
"        table.insert(names, \"execute\")\n"
"    end\n"
"    return table.concat(names, \",\")\n"
"end\n"
"\n"
"local function object_resolve(object_type, object_name)\n"
"    if object_type == 'universe' then\n"
"        return 0\n"
"    end\n"
"    if object_type == 'space' then\n"
"        local space = box.space[object_name]\n"
"        if  space == nil then\n"
"            box.error(box.error.NO_SUCH_SPACE, object_name)\n"
"        end\n"
"        return space.id\n"
"    end\n"
"    if object_type == 'function' then\n"
"        local _func = box.space[box.schema.FUNC_ID]\n"
"        local func\n"
"        if type(object_name) == 'string' then\n"
"            func = _func.index.name:get{object_name}\n"
"        else\n"
"            func = _func:get{object_name}\n"
"        end\n"
"        if func then\n"
"            return func[1]\n"
"        else\n"
"            box.error(box.error.NO_SUCH_FUNCTION, object_name)\n"
"        end\n"
"    end\n"
"    if object_type == 'role' then\n"
"        local _user = box.space[box.schema.USER_ID]\n"
"        local role\n"
"        if type(object_name) == 'string' then\n"
"            role = _user.index.name:get{object_name}\n"
"        else\n"
"            role = _user:get{object_name}\n"
"        end\n"
"        if role and role[4] == 'role' then\n"
"            return role[1]\n"
"        else\n"
"            box.error(box.error.NO_SUCH_ROLE, object_name)\n"
"        end\n"
"    end\n"
"\n"
"    box.error(box.error.UNKNOWN_SCHEMA_OBJECT, object_type)\n"
"end\n"
"\n"
"local function object_name(object_type, object_id)\n"
"    if object_type == 'universe' then\n"
"        return \"\"\n"
"    end\n"
"    local space\n"
"    if object_type == 'space' then\n"
"        space = box.space._space\n"
"    elseif object_type == 'function' then\n"
"        space = box.space._func\n"
"    elseif object_type == 'role' or object_type == 'user' then\n"
"        space = box.space._user\n"
"    else\n"
"        box.error(box.error.UNKNOWN_SCHEMA_OBJECT, object_type)\n"
"    end\n"
"    return space:get{object_id}[3]\n"
"end\n"
"\n"
"box.schema.func = {}\n"
"box.schema.func.create = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { setuid = 'boolean',\n"
"                              if_not_exists = 'boolean',\n"
"                              language = 'string'})\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local func = _func.index.name:get{name}\n"
"    if func then\n"
"        if not opts.if_not_exists then\n"
"            box.error(box.error.FUNCTION_EXISTS, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    opts = update_param_table(opts, { setuid = false, language = 'lua'})\n"
"    opts.language = string.upper(opts.language)\n"
"    opts.setuid = opts.setuid and 1 or 0\n"
"    _func:auto_increment{session.uid(), name, opts.setuid, opts.language}\n"
"end\n"
"\n"
"box.schema.func.drop = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_exists = 'boolean' })\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local fid\n"
"    local tuple\n"
"    if type(name) == 'string' then\n"
"        tuple = _func.index.name:get{name}\n"
"    else\n"
"        tuple = _func:get{name}\n"
"    end\n"
"    if tuple then\n"
"        fid = tuple[1]\n"
"    end\n"
"    if fid == nil then\n"
"        if not opts.if_exists then\n"
"            box.error(box.error.NO_SUCH_FUNCTION, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    local privs = _priv.index.object:select{'function', fid}\n"
"    for k, tuple in pairs(privs) do\n"
"        box.schema.user.revoke(tuple[2], tuple[5], tuple[3], tuple[4])\n"
"    end\n"
"    _func:delete{fid}\n"
"end\n"
"\n"
"function box.schema.func.exists(name_or_id)\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local tuple = nil\n"
"    if type(name_or_id) == 'string' then\n"
"        tuple = _func.index.name:get{name_or_id}\n"
"    elseif type(name_or_id) == 'number' then\n"
"        tuple = _func:get{name_or_id}\n"
"    end\n"
"    return tuple ~= nil\n"
"end\n"
"\n"
"box.schema.func.reload = internal.func_reload\n"
"\n"
"box.schema.user = {}\n"
"\n"
"box.schema.user.password = function(password)\n"
"    local BUF_SIZE = 128\n"
"    local buf = ffi.new(\"char[?]\", BUF_SIZE)\n"
"    builtin.password_prepare(password, #password, buf, BUF_SIZE)\n"
"    return ffi.string(buf)\n"
"end\n"
"\n"
"local function chpasswd(uid, new_password)\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    local auth_mech_list = {}\n"
"    auth_mech_list[\"chap-sha1\"] = box.schema.user.password(new_password)\n"
"    _user:update({uid}, {{\"=\", 5, auth_mech_list}})\n"
"end\n"
"\n"
"box.schema.user.passwd = function(name, new_password)\n"
"    if name == nil then\n"
"        box.error(box.error.PROC_LUA, \"Usage: box.schema.user.passwd([user,] password)\")\n"
"    end\n"
"    if new_password == nil then\n"
"        -- change password for current user\n"
"        new_password = name\n"
"        box.session.su('admin', chpasswd, session.uid(), new_password)\n"
"    else\n"
"        -- change password for other user\n"
"        local uid = user_resolve(name)\n"
"        if uid == nil then\n"
"            box.error(box.error.NO_SUCH_USER, name)\n"
"        end\n"
"        return chpasswd(uid, new_password)\n"
"    end\n"
"end\n"
"\n"
"box.schema.user.create = function(name, opts)\n"
"    local uid = user_or_role_resolve(name)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { password = 'string', if_not_exists = 'boolean' })\n"
"    if uid then\n"
"        if not opts.if_not_exists then\n"
"            box.error(box.error.USER_EXISTS, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    local auth_mech_list = setmap({})\n"
"    if opts.password then\n"
"        auth_mech_list[\"chap-sha1\"] = box.schema.user.password(opts.password)\n"
"    end\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    uid = _user:auto_increment{session.uid(), name, 'user', auth_mech_list}[1]\n"
"    -- grant role 'public' to the user\n"
"    box.schema.user.grant(uid, 'public')\n"
"end\n"
"\n"
"box.schema.user.exists = function(name)\n"
"    if user_resolve(name) then\n"
"        return true\n"
"    else\n"
"        return false\n"
"    end\n"
"end\n"
"\n"
"local function grant(uid, name, privilege, object_type,\n"
"                     object_name, options)\n"
"    -- From user point of view, role is the same thing\n"
"    -- as a privilege. Allow syntax grant(user, role).\n"
"    if object_name == nil and object_type == nil then\n"
"        -- sic: avoid recursion, to not bother with roles\n"
"        -- named 'execute'\n"
"        object_type = 'role'\n"
"        object_name = privilege\n"
"        privilege = 'execute'\n"
"    end\n"
"    local privilege_hex = checked_privilege(privilege, object_type)\n"
"\n"
"    local oid = object_resolve(object_type, object_name)\n"
"    options = options or {}\n"
"    if options.grantor == nil then\n"
"        options.grantor = session.uid()\n"
"    else\n"
"        options.grantor = user_or_role_resolve(options.grantor)\n"
"    end\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    -- add the granted privilege to the current set\n"
"    local tuple = _priv:get{uid, object_type, oid}\n"
"    local old_privilege\n"
"    if tuple ~= nil then\n"
"        old_privilege = tuple[5]\n"
"    else\n"
"        old_privilege = 0\n"
"    end\n"
"    privilege_hex = bit.bor(privilege_hex, old_privilege)\n"
"    -- do not execute a replace if it does not change anything\n"
"    -- XXX bug if we decide to add a grant option: new grantor\n"
"    -- replaces the old one, old grantor is lost\n"
"    if privilege_hex ~= old_privilege then\n"
"        _priv:replace{options.grantor, uid, object_type, oid, privilege_hex}\n"
"    elseif not options.if_not_exists then\n"
"            if object_type == 'role' then\n"
"                box.error(box.error.ROLE_GRANTED, name, object_name)\n"
"            else\n"
"                box.error(box.error.PRIV_GRANTED, name, privilege,\n"
"                          object_type, object_name)\n"
"            end\n"
"    end\n"
"end\n"
"\n"
"local function revoke(uid, name, privilege, object_type, object_name, options)\n"
"    -- From user point of view, role is the same thing\n"
"    -- as a privilege. Allow syntax revoke(user, role).\n"
"    if object_name == nil and object_type == nil then\n"
"        object_type = 'role'\n"
"        object_name = privilege\n"
"        privilege = 'execute'\n"
"    end\n"
"    local privilege_hex = checked_privilege(privilege, object_type)\n"
"\n"
"    options = options or {}\n"
"    local oid = object_resolve(object_type, object_name)\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local tuple = _priv:get{uid, object_type, oid}\n"
"    if tuple == nil then\n"
"        if options.if_exists then\n"
"            return\n"
"        end\n"
"        if object_type == 'role' then\n"
"            box.error(box.error.ROLE_NOT_GRANTED, name, object_name)\n"
"        else\n"
"            box.error(box.error.PRIV_NOT_GRANTED, name, privilege,\n"
"                      object_type, object_name)\n"
"        end\n"
"    end\n"
"    local old_privilege = tuple[5]\n"
"    local grantor = tuple[1]\n"
"    -- sic:\n"
"    -- a user may revoke more than he/she granted\n"
"    -- (erroneous user input)\n"
"    --\n"
"    privilege_hex = bit.band(old_privilege, bit.bnot(privilege_hex))\n"
"    if privilege_hex ~= 0 then\n"
"        _priv:replace{grantor, uid, object_type, oid, privilege_hex}\n"
"    else\n"
"        _priv:delete{uid, object_type, oid}\n"
"    end\n"
"end\n"
"\n"
"local function drop(uid, opts)\n"
"    -- recursive delete of user data\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local privs = _priv.index.primary:select{uid}\n"
"    for k, tuple in pairs(privs) do\n"
"        revoke(uid, uid, tuple[5], tuple[3], tuple[4])\n"
"    end\n"
"    local spaces = box.space[box.schema.SPACE_ID].index.owner:select{uid}\n"
"    for k, tuple in pairs(spaces) do\n"
"        box.space[tuple[1]]:drop()\n"
"    end\n"
"    local funcs = box.space[box.schema.FUNC_ID].index.owner:select{uid}\n"
"    for k, tuple in pairs(funcs) do\n"
"        box.schema.func.drop(tuple[1])\n"
"    end\n"
"    -- if this is a role, revoke this role from whoever it was granted to\n"
"    local grants = _priv.index.object:select{'role', uid}\n"
"    for k, tuple in pairs(grants) do\n"
"        revoke(tuple[2], tuple[2], uid)\n"
"    end\n"
"    box.space[box.schema.USER_ID]:delete{uid}\n"
"end\n"
"\n"
"box.schema.user.grant = function(user_name, ...)\n"
"    local uid = user_resolve(user_name)\n"
"    if uid == nil then\n"
"        box.error(box.error.NO_SUCH_USER, user_name)\n"
"    end\n"
"    return grant(uid, user_name, ...)\n"
"end\n"
"\n"
"box.schema.user.revoke = function(user_name, ...)\n"
"    local uid = user_resolve(user_name)\n"
"    if uid == nil then\n"
"        box.error(box.error.NO_SUCH_USER, user_name)\n"
"    end\n"
"    return revoke(uid, user_name, ...)\n"
"end\n"
"\n"
"box.schema.user.drop = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_exists = 'boolean' })\n"
"    local uid = user_resolve(name)\n"
"    if uid ~= nil then\n"
"        if uid >= box.schema.SYSTEM_USER_ID_MIN and\n"
"           uid <= box.schema.SYSTEM_USER_ID_MAX then\n"
"            -- gh-1205: box.schema.user.info fails\n"
"            box.error(box.error.DROP_USER, name,\n"
"                      \"the user or the role is a system\")\n"
"        end\n"
"        return drop(uid, opts)\n"
"    end\n"
"    if not opts.if_exists then\n"
"        box.error(box.error.NO_SUCH_USER, name)\n"
"    end\n"
"    return\n"
"end\n"
"\n"
"local function info(id)\n"
"    local _priv = box.space._priv\n"
"    local _user = box.space._priv\n"
"    local privs = {}\n"
"    for _, v in pairs(_priv:select{id}) do\n"
"        table.insert(\n"
"            privs,\n"
"            {privilege_name(v[5]), v[3], object_name(v[3], v[4])}\n"
"        )\n"
"    end\n"
"    return privs\n"
"end\n"
"\n"
"box.schema.user.info = function(user_name)\n"
"    local uid\n"
"    if user_name == nil then\n"
"        uid = box.session.uid()\n"
"    else\n"
"        uid = user_resolve(user_name)\n"
"        if uid == nil then\n"
"            box.error(box.error.NO_SUCH_USER, user_name)\n"
"        end\n"
"    end\n"
"    return info(uid)\n"
"end\n"
"\n"
"box.schema.role = {}\n"
"\n"
"box.schema.role.exists = function(name)\n"
"    if role_resolve(name) then\n"
"        return true\n"
"    else\n"
"        return false\n"
"    end\n"
"end\n"
"\n"
"box.schema.role.create = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_not_exists = 'boolean' })\n"
"    local uid = user_or_role_resolve(name)\n"
"    if uid then\n"
"        if not opts.if_not_exists then\n"
"            box.error(box.error.ROLE_EXISTS, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    _user:auto_increment{session.uid(), name, 'role', setmap({})}\n"
"end\n"
"\n"
"box.schema.role.drop = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_exists = 'boolean' })\n"
"    local uid = role_resolve(name)\n"
"    if uid == nil then\n"
"        if not opts.if_exists then\n"
"            box.error(box.error.NO_SUCH_ROLE, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    if uid >= box.schema.SYSTEM_USER_ID_MIN and\n"
"       uid <= box.schema.SYSTEM_USER_ID_MAX then\n"
"        -- gh-1205: box.schema.user.info fails\n"
"        box.error(box.error.DROP_USER, name, \"the user or the role is a system\")\n"
"    end\n"
"    return drop(uid)\n"
"end\n"
"box.schema.role.grant = function(user_name, ...)\n"
"    local uid = role_resolve(user_name)\n"
"    if uid == nil then\n"
"        box.error(box.error.NO_SUCH_ROLE, user_name)\n"
"    end\n"
"    return grant(uid, user_name, ...)\n"
"end\n"
"box.schema.role.revoke = function(user_name, ...)\n"
"    local uid = role_resolve(user_name)\n"
"    if uid == nil then\n"
"        box.error(box.error.NO_SUCH_ROLE, user_name)\n"
"    end\n"
"    return revoke(uid, user_name, ...)\n"
"end\n"
"box.schema.role.info = function(role_name)\n"
"    local rid = role_resolve(role_name)\n"
"    if rid == nil then\n"
"        box.error(box.error.NO_SUCH_ROLE, role_name)\n"
"    end\n"
"    return info(rid)\n"
"end\n"
"\n"
"--\n"
"-- once\n"
"--\n"
"box.once = function(key, func, ...)\n"
"    if type(key) ~= 'string' or type(func) ~= 'function' then\n"
"        box.error(box.error.ILLEGAL_PARAMS, \"Usage: box.once(key, func, ...)\")\n"
"    end\n"
"\n"
"    local key = \"once\"..key\n"
"    if box.space._schema:get{key} ~= nil then\n"
"        return\n"
"    end\n"
"    box.space._schema:put{key}\n"
"    return func(...)\n"
"end\n"
"\n"
"--\n"
"-- nice output when typing box.space in admin console\n"
"--\n"
"box.space = {}\n"
"\n"
"local function box_space_mt(tab)\n"
"    local t = {}\n"
"    for k,v in pairs(tab) do\n"
"        -- skip system spaces and views\n"
"        if type(k) == 'string' and #k > 0 and k:sub(1,1) ~= '_' then\n"
"            t[k] = { engine = v.engine, temporary = v.temporary }\n"
"        end\n"
"    end\n"
"    return t\n"
"end\n"
"\n"
"setmetatable(box.space, { __serialize = box_space_mt })\n"
""
;
