const char argparse_lua[] =
"local fun = require('fun')\n"
"\n"
"local function parse_param_prefix(param)\n"
"    if param == nil then return nil end\n"
"    local is_long  = (param:find(\"^[-][-]\") ~= nil)\n"
"    local is_short = not is_long and (param:find(\"^[-]\") ~= nil)\n"
"    local is_dash  = is_short and (param:find(\"^[-]$\") ~= nil)\n"
"    return is_long, is_short, is_dash\n"
"end\n"
"\n"
"local function result_set_add(t_out, key, val)\n"
"    if val == nil then\n"
"        table.insert(t_out, key)\n"
"    elseif t_out[key] == nil then\n"
"        t_out[key] = val\n"
"    elseif type(t_out[key]) == 'table' then\n"
"        table.insert(t_out[key], val)\n"
"    else\n"
"        t_out[key] = {t_out[key], val}\n"
"    end\n"
"end\n"
"\n"
"local function convert_parameter_simple(name, convert_from, convert_to)\n"
"    if convert_to == 'number' then\n"
"        local converted = tonumber(convert_from)\n"
"        if converted == nil then\n"
"            error(\n"
"                ('Bad value for parameter %s. expected type %s, got \"%s\"')\n"
"                :format(name, convert_to, convert_from)\n"
"            )\n"
"        end\n"
"        return converted\n"
"    elseif convert_to == 'boolean' then\n"
"        if type(convert_from) ~= 'boolean' then\n"
"            error(\n"
"                ('Bad input for parameter \"%s\". Expected boolean, got \"%s\"')\n"
"                :format(name, convert_from)\n"
"            )\n"
"        end\n"
"    elseif convert_to == 'string' then\n"
"        if type(convert_from) ~= 'string' then\n"
"            error(\n"
"                ('Bad input for parameter \"%s\". Expected string, got \"%s\"')\n"
"                :format(name, convert_from)\n"
"            )\n"
"        end\n"
"    else\n"
"        error(\n"
"            ('Bad convertion format \"%s\" provided for %s')\n"
"            :format(convert_to, name)\n"
"        )\n"
"    end\n"
"    return convert_from\n"
"end\n"
"\n"
"local function convert_parameter(name, convert_from, convert_to)\n"
"    if convert_to == nil then\n"
"        return convert_from\n"
"    end\n"
"    if convert_to:find('+') then\n"
"        convert_to = convert_to:sub(1, -2)\n"
"        if type(convert_from) ~= 'table' then\n"
"            convert_from = { convert_from }\n"
"        end\n"
"        convert_from = fun.iter(convert_from):map(function(v)\n"
"            return convert_parameter_simple(name, v, convert_to)\n"
"        end):totable()\n"
"    else\n"
"        if type(convert_from) == 'table' then\n"
"            convert_from = table.remove(convert_from)\n"
"        end\n"
"        convert_from = convert_parameter_simple(name, convert_from, convert_to)\n"
"    end\n"
"    return convert_from\n"
"end\n"
"\n"
"local function parameters_parse(t_in, options)\n"
"    local t_out, t_in = {}, t_in or {}\n"
"    local skip_param = false\n"
"    for i, v in ipairs(t_in) do\n"
"        -- we've used this parameter as value\n"
"        if skip_param == true then\n"
"            skip_param = false\n"
"            goto nextparam\n"
"        end\n"
"        local is_long, is_short, is_dash = parse_param_prefix(v)\n"
"        if not is_dash and is_short then\n"
"            local commands = v:sub(2)\n"
"            if not (commands:match(\"^[%a]+$\")) then\n"
"                error((\"bad argument #%d: ID not valid\"):format(i))\n"
"            end\n"
"            for id in v:sub(2):gmatch(\"%a\") do\n"
"                result_set_add(t_out, id, true)\n"
"            end\n"
"        elseif is_long then\n"
"            local command = v:sub(3)\n"
"            if command:find('=') then\n"
"                local key, val = command:match(\"^([%a_][%w_-]+)%=(.*)$\")\n"
"                if key == nil or val == nil then\n"
"                    error((\"bad argument #%d: ID not valid\"):format(i))\n"
"                end\n"
"                result_set_add(t_out, key, val)\n"
"            else\n"
"                if command:match(\"^([%a_][%w_-]+)$\") == nil then\n"
"                    error((\"bad argument #%d: ID not valid\"):format(i))\n"
"                end\n"
"                local val = true\n"
"                do\n"
"                    -- in case next argument is value of this key (not --arg)\n"
"                    local next_arg = t_in[i + 1]\n"
"                    local is_long, is_short, is_dash = parse_param_prefix(next_arg)\n"
"                    if is_dash then\n"
"                        skip_param = true\n"
"                    elseif is_long == false and not is_short and not is_dash then\n"
"                        val = next_arg\n"
"                        skip_param = true\n"
"                    end\n"
"                end\n"
"                result_set_add(t_out, command, val)\n"
"            end\n"
"        else\n"
"            table.insert(t_out, v)\n"
"        end\n"
"::nextparam::\n"
"    end\n"
"    if options then\n"
"        local lookup, unknown = {}, {}\n"
"        for _, v in ipairs(options) do\n"
"            if type(v) ~= 'table' then\n"
"                v = {v}\n"
"            end\n"
"            lookup[v[1]] = (v[2] or true)\n"
"        end\n"
"        for k, v in pairs(t_out) do\n"
"            if lookup[k] == nil and type(k) == \"string\" then\n"
"                table.insert(unknown, k)\n"
"            elseif type(lookup[k]) == 'string' then\n"
"                t_out[k] = convert_parameter(k, v, lookup[k])\n"
"            end\n"
"        end\n"
"        if #unknown > 0 then\n"
"            error((\"unknown options: %s\"):format(table.concat(unknown, \", \")))\n"
"        end\n"
"    end\n"
"    return t_out\n"
"end\n"
"\n"
"return {\n"
"    parse = parameters_parse\n"
"}\n"
""
;
