const char log_lua[] =
"-- log.lua\n"
"--\n"
"local ffi = require('ffi')\n"
"ffi.cdef[[\n"
"    typedef void (*sayfunc_t)(int level, const char *filename, int line,\n"
"               const char *error, const char *format, ...);\n"
"    void\n"
"    say_set_log_level(int new_level);\n"
"\n"
"    extern sayfunc_t _say;\n"
"    extern void say_logrotate(int);\n"
"\n"
"    enum say_level {\n"
"        S_FATAL,\n"
"        S_SYSERROR,\n"
"        S_ERROR,\n"
"        S_CRIT,\n"
"        S_WARN,\n"
"        S_INFO,\n"
"        S_VERBOSE,\n"
"        S_DEBUG\n"
"    };\n"
"\n"
"    pid_t log_pid;\n"
"    extern int log_level;\n"
"]]\n"
"\n"
"local S_WARN = ffi.C.S_WARN\n"
"local S_INFO = ffi.C.S_INFO\n"
"local S_VERBOSE = ffi.C.S_VERBOSE\n"
"local S_DEBUG = ffi.C.S_DEBUG\n"
"local S_ERROR = ffi.C.S_ERROR\n"
"\n"
"local function say(level, fmt, ...)\n"
"    if ffi.C.log_level < level then\n"
"        -- don't waste cycles on debug.getinfo()\n"
"        return\n"
"    end\n"
"    local debug = require('debug')\n"
"    if select('#', ...) ~= 0 then\n"
"        local stat\n"
"        stat, fmt = pcall(string.format, fmt, ...)\n"
"        if not stat then\n"
"            error(fmt, 3)\n"
"        end\n"
"    elseif type(fmt) ~= 'string' then\n"
"        fmt = tostring(fmt)\n"
"    end\n"
"    local frame = debug.getinfo(3, \"Sl\")\n"
"    local line, file = 0, 'eval'\n"
"    if type(frame) == 'table' then\n"
"        line = frame.currentline or 0\n"
"        file = frame.short_src or frame.src or 'eval'\n"
"    end\n"
"    ffi.C._say(level, file, line, nil, \"%s\", fmt)\n"
"end\n"
"\n"
"local function say_closure(lvl)\n"
"    return function (fmt, ...)\n"
"        say(lvl, fmt, ...)\n"
"    end\n"
"end\n"
"\n"
"local function log_rotate()\n"
"    ffi.C.say_logrotate(0)\n"
"end\n"
"\n"
"local function log_level(level)\n"
"    return ffi.C.say_set_log_level(level)\n"
"end\n"
"\n"
"local function log_pid()\n"
"    return tonumber(ffi.C.log_pid)\n"
"end\n"
"\n"
"local compat_warning_said = false\n"
"local compat_v16 = {\n"
"    logger_pid = function()\n"
"        if not compat_warning_said then\n"
"            compat_warning_said = true\n"
"            say(S_WARN, 'logger_pid() is deprecated, please use pid() instead')\n"
"        end\n"
"        return log_pid()\n"
"    end;\n"
"}\n"
"\n"
"return setmetatable({\n"
"    warn = say_closure(S_WARN);\n"
"    info = say_closure(S_INFO);\n"
"    verbose = say_closure(S_VERBOSE);\n"
"    debug = say_closure(S_DEBUG);\n"
"    error = say_closure(S_ERROR);\n"
"    rotate = log_rotate;\n"
"    pid = log_pid;\n"
"    level = log_level;\n"
"}, {\n"
"    __index = compat_v16;\n"
"})\n"
""
;
